/*

Mike DeFehr Sept 2009
mike@defehr.com
http://mikedefehr.com
http://winnipeg.sqlpass.org/

*/

--Simple row number:

USE tempdb
GO
IF OBJECT_ID('DailySales') IS NOT NULL DROP TABLE DailySales
GO
CREATE TABLE DailySales
(	SalesDate	date PRIMARY KEY,
	TotalSales	money
)
GO
INSERT DailySales (SalesDate,TotalSales) VALUES
('2008-08-31',1024),
('2008-09-01',1156),
('2008-09-02',968),
('2008-09-03',1045),
('2008-09-04',1024),
('2008-09-07',1136),
('2008-09-08',954),
('2008-09-09',1049),
('2008-09-10',912),
('2008-09-11',1127),
('2008-09-14',1078),
('2008-09-15',945),
('2008-09-16',1073),
('2008-09-17',1075),
('2008-09-18',1086),
('2008-09-21',945),
('2008-09-22',994),
('2008-09-23',1005),
('2008-09-24',1045),
('2008-09-25',989)
GO

SELECT	SalesDate,
		TotalSales,
		ROW_NUMBER() OVER (ORDER BY (SELECT 0)) AS RowNum
FROM DailySales;









--running sum using row number

WITH ds AS
(
	SELECT	SalesDate,
			TotalSales,
			ROW_NUMBER() OVER (ORDER BY SalesDate) AS RowNum
	FROM DailySales
),
cte AS
(
	SELECT	SalesDate,
			TotalSales,
			TotalSales AS SalesToDate,
			RowNum
	FROM ds WHERE RowNum = 1
	UNION ALL
	SELECT	ds.SalesDate,
			ds.TotalSales,
			ds.TotalSales + cte.SalesToDate,
			ds.RowNum
	FROM ds
		JOIN cte ON cte.RowNum + 1 = ds.RowNum
)
SELECT * FROM cte;	


--Windowed aggregation

--2005+
SELECT *, SUM(TotalSales) OVER()
FROM DailySales

SELECT *, 100 * TotalSales / SUM(TotalSales) OVER() AS PercentOfSales
FROM DailySales
WHERE SalesDate >= '2008-09-01'
	AND SalesDate < '2008-10-01'


--2012+
SELECT *, SUM(TotalSales) OVER(ORDER BY SalesDate)
FROM DailySales

SELECT *, SUM(TotalSales) OVER(ORDER BY SalesDate ROWS BETWEEN UNBOUNDED PRECEDING AND CURRENT ROW)
FROM DailySales

SELECT *, SUM(TotalSales) OVER(ORDER BY SalesDate ROWS BETWEEN 5 PRECEDING AND CURRENT ROW)
FROM DailySales

SELECT *, AVG(TotalSales) OVER(ORDER BY SalesDate ROWS BETWEEN 5 PRECEDING AND CURRENT ROW)
FROM DailySales








--finding groups of consecutive values

SELECT * FROM DailySales;


DECLARE @BaseDate DATE = (SELECT TOP 1 SalesDate FROM DailySales ORDER BY SalesDate);
WITH cte AS
(
	SELECT	SalesDate,
			TotalSales,
			DATEDIFF(DAY,@BaseDate,SalesDate) + 1 AS ConsecutiveDayDiff,
			ROW_NUMBER() OVER (ORDER BY SalesDate) AS RowNumber,
			DATEDIFF(DAY,@BaseDate,SalesDate) + 1 - ROW_NUMBER() OVER (ORDER BY SalesDate) AS Grp
	FROM DailySales	
)
--SELECT * FROM cte
SELECT	ROW_NUMBER() OVER (ORDER BY MIN(SalesDate)) AS GroupNumber,
		MIN(SalesDate) AS DateFrom,
		MAX(SalesDate) AS DateTo,
		COUNT(*) AS NumRecords
FROM cte
GROUP BY grp;

WITH cte AS
(
SELECT	SalesDate, 
		LEAD(SalesDate) OVER(ORDER BY SalesDate) AS NextSalesDate,
		DATEDIFF(day,SalesDate,LEAD(SalesDate) OVER(ORDER BY SalesDate)) - 1 AS DaysSkipped
FROM DailySales
)
SELECT *
FROM cte WHERE DaysSkipped <> 0
GO

DECLARE @BaseDate DATE = (SELECT TOP 1 SalesDate FROM DailySales ORDER BY SalesDate);
WITH cte AS
(
	SELECT	SalesDate,
			TotalSales,
			DATEDIFF(DAY,@BaseDate,SalesDate) + 1 AS ConsecutiveDayDiff,
			ROW_NUMBER() OVER (ORDER BY SalesDate) AS RowNumber,
			DATEDIFF(DAY,@BaseDate,SalesDate) + 1 - ROW_NUMBER() OVER (ORDER BY SalesDate) AS Grp,
			LEAD(SalesDate) OVER(ORDER BY SalesDate) AS NextSalesDate,
			DATEDIFF(day,SalesDate,LEAD(SalesDate) OVER(ORDER BY SalesDate)) - 1 AS DaysSkipped
	FROM DailySales	
)
SELECT	ROW_NUMBER() OVER (ORDER BY MIN(SalesDate)) AS GroupNumber,
		MIN(SalesDate) AS DateFrom,
		MAX(SalesDate) AS DateTo,
		COUNT(*) AS NumConsecutiveDays,
		MAX(DaysSkipped) AS DaysSkipped
FROM cte
GROUP BY grp;


--paging

DECLARE @Page INT = 3;
DECLARE @PageSize INT = 5;
WITH cte AS
(
	SELECT	SalesDate,
			TotalSales,
			--ROW_NUMBER() OVER (ORDER BY SalesDate) / CAST(@PageSize AS FLOAT) AS [Div],
			--ROW_NUMBER() OVER (ORDER BY SalesDate) / @PageSize AS [Div2],
			(ROW_NUMBER() OVER (ORDER BY SalesDate) - 1) / @PageSize + 1 AS Page
	FROM DailySales	
)
SELECT * FROM cte
WHERE Page = @Page;
GO

DECLARE @Page INT = 3;
DECLARE @PageSize INT = 5;
SELECT TOP (@PageSize) SalesDate, TotalSales
FROM (	SELECT	SalesDate,
				TotalSales,
				ROW_NUMBER() OVER (ORDER BY SalesDate) AS rn
		FROM DailySales	) s
WHERE s.rn > (@Page - 1) * @PageSize
GO

DECLARE @Page INT = 3;
DECLARE @PageSize INT = 5;
WITH cte AS
(
	SELECT *, ROW_NUMBER() OVER (ORDER BY SalesDate) AS rn
	FROM DailySales
)
SELECT SalesDate, TotalSales
FROM cte
WHERE rn BETWEEN (@Page - 1) * @PageSize + 1 AND @Page * @PageSize;
GO

--2012+
DECLARE @Page INT = 3;
DECLARE @PageSize INT = 5;
SELECT *
FROM DailySales
ORDER BY SalesDate
OFFSET (@Page-1) * @PageSize ROWS
FETCH NEXT @PageSize ROWS ONLY;





--conditional paging
--http://beyondrelational.com/blogs/tc/

IF OBJECT_ID('Invoices') IS NOT NULL DROP TABLE Invoices
CREATE TABLE Invoices (
	InvID INT IDENTITY,
	BatchNumber INT,
	InvoiceNumber INT, 
	VisitDate DATETIME,
	Amount MONEY )
INSERT Invoices(BatchNumber, InvoiceNumber, VisitDate, Amount) VALUES
( 10000001,20001,'2009-01-01',125.00 ),
( 10000001,20001,'2009-01-02',50.00 ),
( 10000001,20001,'2009-01-03',150.00 ),
( 10000001,20001,'2009-01-04',75.00 ),
( 10000001,20002,'2009-01-01',175.00 ),
( 10000001,20002,'2009-01-02',200.00 ),
( 10000001,20002,'2009-01-03',250.00 ),
( 10000001,20003,'2009-01-01',85.00 ),
( 10000001,20003,'2009-01-02',105.00 ),
( 10000001,20004,'2009-01-01',135.00 ),
( 10000001,20004,'2009-01-02',95.00 ),
( 10000001,20004,'2009-01-03',110.00 ),
( 10000001,20004,'2009-01-04',55.00 ),
( 10000001,20004,'2009-01-05',130.00 ),
( 10000001,20004,'2009-01-06',180.00 ),
( 10000001,20005,'2009-01-01',185.00 ),
( 10000001,20005,'2009-01-02',145.00 ),
( 10000001,20005,'2009-01-13',205.00 ),
( 10000001,20005,'2009-01-14',65.00 ),
( 10000001,20005,'2009-01-15',85.00 ),
( 10000001,20005,'2009-01-06',285.00 ),
( 10000001,20005,'2009-01-07',210.00 ),
( 10000001,20005,'2009-01-08',195.00 ),
( 10000001,20006,'2009-01-01',175.00 ),
( 10000002,20007,'2009-01-01',165.00 ),
( 10000002,20007,'2009-01-02',65.00 ),
( 10000002,20007,'2009-01-03',175.00 ),
( 10000002,20008,'2009-01-01',195.00 ),
( 10000002,20008,'2009-01-02',125.00 ),
( 10000002,20008,'2009-01-03',185.00 );

WITH subcount AS
(	--create a grouped cte to provide rowcounts for all invoices
SELECT	BatchNumber, 
		InvoiceNumber, 
		COUNT(*) AS InvCount, 
		ROW_NUMBER() OVER (ORDER BY InvoiceNumber) AS RowNum
FROM Invoices
GROUP BY BatchNumber, InvoiceNumber
),
cte AS
(	--use a recursive cte to iterate through the grouped list 
	--and add do a running sum of the invoice totals, 
	--conditionally resetting the total when a new batch or a 
	--total of 10 is reached
SELECT	BatchNumber, 
		InvoiceNumber, 
		InvCount, 
		InvCount AS RunSum,
		1 AS [Page], 
		RowNum
FROM subcount
WHERE RowNum = 1
UNION ALL
SELECT	sc.BatchNumber,
		sc.InvoiceNumber,
		sc.InvCount,
		CASE WHEN sc.BatchNumber <> cte.BatchNumber OR sc.InvCount + cte.RunSum > 10 
			THEN sc.InvCount	--reset case - throw away the running sum and start a new one with the current count
			ELSE sc.InvCount + cte.RunSum --regular case - add the previous record's total to the current count
		END,
		CASE WHEN sc.BatchNumber <> cte.BatchNumber OR sc.InvCount + cte.RunSum > 10 
			THEN	cte.[Page] + 1 --reset case - increment the Page number
			ELSE  cte.[Page] --regular case - repeat the Page number
		END,
		sc.RowNum
FROM subcount sc
	JOIN cte ON sc.RowNum = cte.RowNum + 1
)
--SELECT * FROM cte
SELECT t.*, cte.[Page]
FROM cte
	JOIN Invoices t ON cte.InvoiceNumber = t.InvoiceNumber
ORDER BY cte.BatchNumber, cte.InvoiceNumber








-- working with duplicates

IF OBJECT_ID('DailySales') IS NOT NULL DROP TABLE DailySales
GO
CREATE TABLE DailySales
(	SalesDate	date,
	TotalSales	money
)
GO
INSERT DailySales (SalesDate,TotalSales) VALUES
('2008-09-01',1156),
('2008-09-02',968),
('2008-09-21',945),
('2008-09-03',1045),
('2008-09-04',1024),
('2008-09-07',1136),
('2008-09-08',954),
('2008-09-09',1049),
('2008-09-10',912),
('2008-09-03',1045),
('2008-09-11',1127),
('2008-09-14',1078),
('2008-09-15',945),
('2008-09-16',1073),
('2008-09-17',1075),
('2008-09-18',1086),
('2008-09-21',945),
('2008-09-22',994),
('2008-09-23',1005),
('2008-09-10',912),
('2008-09-24',1045)
GO


SELECT * 
FROM DailySales
ORDER BY SalesDate

--we need to find the duplicates

SELECT SalesDate, COUNT(*) AS DupCount
FROM DailySales
GROUP BY SalesDate
HAVING COUNT(*) > 1;


WITH cte AS
(
	SELECT	SalesDate,
			TotalSales,
			COUNT(*) OVER(PARTITION BY SalesDate) AS DateCount
	FROM DailySales
)
SELECT * FROM cte
WHERE DateCount > 1;

--what about deleting the duplicates?

WITH cte AS
(
	SELECT	SalesDate,
			TotalSales,
			ROW_NUMBER() OVER(PARTITION BY SalesDate ORDER BY TotalSales) AS DateRowNum
	FROM DailySales
)
--SELECT * FROM cte
DELETE FROM cte WHERE DateRowNum > 1







--TOP N by category

IF OBJECT_ID('DailySales') IS NOT NULL DROP TABLE DailySales
GO
CREATE TABLE DailySales
(	SalesDate	date PRIMARY KEY,
	TotalSales	money
)
GO

WITH nums AS
(
	SELECT ROW_NUMBER() OVER (ORDER BY (SELECT 0)) AS num
	FROM sys.all_objects a, sys.all_objects b
)
INSERT DailySales (SalesDate, TotalSales)
SELECT	DATEADD(day,num-1,'2009-01-01'), 
		CAST(RAND(CHECKSUM(NEWID())) * 100.0 AS decimal(19,2)) + 950
FROM nums
WHERE num <= 365
	AND DATEPART(weekday,DATEADD(day,num-1,'2009-01-01')) NOT IN (1,7)

	
SELECT * FROM DailySales;


--Top 3 days in each month

WITH cte AS
(
	SELECT	 SalesDate
			,TotalSales
			,ROW_NUMBER() OVER (PARTITION BY YEAR(SalesDate), MONTH(SalesDate) ORDER BY TotalSales DESC) AS MonthRowNum
			--,RANK() OVER (PARTITION BY YEAR(SalesDate), MONTH(SalesDate) ORDER BY TotalSales DESC) AS MonthRank
			--,DENSE_RANK() OVER (PARTITION BY YEAR(SalesDate), MONTH(SalesDate) ORDER BY TotalSales DESC) AS MonthDenseRank
			--,NTILE(4) OVER (PARTITION BY YEAR(SalesDate), MONTH(SalesDate) ORDER BY TotalSales DESC) AS MonthPercentile
	FROM DailySales
)
SELECT * FROM cte
WHERE MonthRowNum <= 3

UPDATE DailySales SET TotalSales = 1045.58 WHERE SalesDate = '2009-01-27'
