/*

Mike DeFehr Sept 2009
mike@defehr.com
http://mikedefehr.com
http://winnipeg.sqlpass.org/

*/


--Conventional String Splitting

DECLARE @Array varchar(255) = 'ABC,DEF,GHI,JKL', 
		@Del char(1) = ','
DECLARE @OutTable TABLE (Element varchar(255))
DECLARE @start int = 1, @end int
	
SET @end = CHARINDEX(@Del,@Array)

WHILE @end <> 0
BEGIN
	INSERT @OutTable (Element) VALUES (SUBSTRING(@Array,@start,@end-@start))
	SET @start = @end + 1
	SET @end = CHARINDEX(@Del,@Array,@start)
END
INSERT @OutTable (Element) VALUES (SUBSTRING(@Array,@start,LEN(@Array)-@start+1))

SELECT * FROM @OutTable

--Create a function for it

USE tempdb
IF OBJECT_ID('fn_split_conv') IS NOT NULL DROP FUNCTION fn_split_conv
GO
CREATE FUNCTION fn_split_conv(@Array varchar(max), @Del char(1))
RETURNS @OutTable TABLE (Element varchar(100))
AS
BEGIN

	DECLARE @start int = 1, @end int = -1
		
	SET @end = CHARINDEX(@Del,@Array)

	WHILE @end <> 0
	BEGIN
		INSERT @OutTable (Element) VALUES (SUBSTRING(@Array,@start,@end-@start))
		SET @start = @end + 1
		SET @end = CHARINDEX(@Del,@Array,@start)
	END
	INSERT @OutTable (Element) VALUES (SUBSTRING(@Array,@start,LEN(@Array)-@start+1))
	
	RETURN
END
GO
--test function

SELECT * FROM dbo.fn_split_conv('ABC,DEF,GHI,JKL',',')








--XML method

DECLARE @Array nvarchar(max) = 'ABC,DEF,GHI,JKL'
DECLARE @Del char(1) = ','

DECLARE @ArrayXML XML = CAST('<e>' + REPLACE(@Array,@Del,'</e><e>') + '</e>' AS XML);
SELECT CAST(@ArrayXML as varchar(1000))

SELECT ax.element.value('.','varchar(100)') FROM @ArrayXML.nodes('/e') ax(element)


IF OBJECT_ID('fn_split_xml') IS NOT NULL DROP FUNCTION fn_split_xml
GO
CREATE FUNCTION fn_split_xml(@Array varchar(max), @Del char(1))
RETURNS @OutTable TABLE (Element varchar(100))
AS
BEGIN
	DECLARE @ArrayXML XML = CAST('<e>' + REPLACE(@Array,@Del,'</e><e>') + '</e>' AS XML);
	
	INSERT @OutTable (Element)
	SELECT ax.element.value('.','varchar(100)') 
	FROM @ArrayXML.nodes('/e') ax(element)

	RETURN
END
GO

--test function

SELECT * FROM dbo.fn_split_xml('ABC,DEF,GHI,JKL',',')








--unconventional way


--first, create a table of numbers
USE tempdb
GO
IF OBJECT_ID('numlist') IS NOT NULL DROP TABLE numlist
CREATE TABLE numlist (num int NOT NULL PRIMARY KEY)
DECLARE @base int, @MaxNum int
SET @MaxNum = 500000

SET @base = 1

INSERT numlist VALUES (1)

WHILE @base * 2 <= @MaxNum
BEGIN
	INSERT numlist
	SELECT num + @base 
	FROM numlist
	
	SET @base = @base * 2
END

INSERT numlist
SELECT num + @base 
FROM numlist
WHERE num + @base <= @MaxNum
GO

SELECT TOP 100 * FROM numlist

DECLARE @Array varchar(max) = 'abcd,efgh,ijkl,mnop,qrst,uvwx,yz', 
		@Del char(1) = ','

SELECT	num
		,SUBSTRING(@Array,num,1) AS CurChar
		--,SUBSTRING(@Array,num,CHARINDEX(@Del,@Array+@Del,num)-num) AS RowData
		--,SUBSTRING(@Del+@Array, num, 1) AS RowCriteria
		--,(num-1) - LEN(REPLACE(LEFT(@Array,num-1),@Del,'')) + 1 AS RowNumber
FROM numlist
WHERE num <= LEN(@Array) + 1
	--AND SUBSTRING(@Del+ @Array, num, 1) = @Del


IF OBJECT_ID('fn_split_big') IS NOT NULL DROP FUNCTION fn_split_big
GO
CREATE FUNCTION fn_split_big(@Array varchar(max), @Del char(1))
RETURNS TABLE
AS
RETURN
	SELECT	SUBSTRING(@Array,num,CHARINDEX(@Del,@Array+@Del,num)-num) AS Element
	FROM numlist
	WHERE num <= LEN(@Array) + 1
		AND SUBSTRING(@Del+ @Array, num, 1) = @Del
GO


--test function

SELECT * FROM dbo.fn_split_big('ABC,DEF,GHI,JKL',',')






--recursive cte method

DECLARE @Array nvarchar(max) = 'ABC,DEF,GHI,JKL'
DECLARE @Del char(1) = ','

SELECT @Array;

WITH cte AS 
(
	SELECT CAST(1 as bigint) AS StrStart,
		CHARINDEX(@Del, @Array + @Del) AS StrStop
	UNION ALL
	SELECT StrStop + 1,
		CHARINDEX(@Del, @Array + @Del, StrStop + 1)
	FROM   cte
	WHERE  StrStop > 0
)
SELECT *, SUBSTRING(@Array, StrStart, StrStop - StrStart) AS Element
FROM   cte
WHERE  StrStop > 0
OPTION (MAXRECURSION 0)




IF OBJECT_ID('fn_split_cte') IS NOT NULL DROP FUNCTION fn_split_cte
GO
CREATE FUNCTION fn_split_cte(@Array varchar(max), @Del char(1))
RETURNS @OutTable TABLE (Element varchar(100))
AS
BEGIN
	WITH cte AS 
	(
		SELECT CAST(1 as bigint) AS StrStart,
			CHARINDEX(@Del, @Array + @Del) AS StrStop
		UNION ALL
		SELECT StrStop + 1,
			CHARINDEX(@Del, @Array + @Del, StrStop+ 1)
		FROM   cte
		WHERE  StrStop > 0
	)
	INSERT @OutTable
	SELECT SUBSTRING(@Array, StrStart, CASE WHEN StrStop > 0 THEN StrStop - StrStart ELSE 0 END) AS Element
	FROM   cte
	WHERE  StrStop > 0
	OPTION (MAXRECURSION 0)
  
	RETURN
 END
GO



SELECT * FROM dbo.fn_split_cte('ABC,DEF,GHI,JKL',',')




--some code to test longer strings (warning: some of these will spin for awhile)

DECLARE @Array nvarchar(max), @NumElements int, @Start datetime
SELECT @Array = SUBSTRING((
SELECT ',' + CAST(num AS varchar) AS [text()]
FROM numlist
WHERE num <= 100000
FOR XML PATH('')),2,9999999)

PRINT @Array

PRINT LEN(@Array)

SET @Start = GETDATE()
SELECT @NumElements = COUNT(*) FROM fn_split_conv(@Array,',')
PRINT 'Conventional: ' + CAST(DATEDIFF(second,@Start,GETDATE()) AS varchar)
SET @Start = GETDATE()
SELECT @NumElements = COUNT(*) FROM fn_split_xml(@Array,' ')
PRINT 'XML: ' + CAST(DATEDIFF(second,@Start,GETDATE()) AS varchar)
SET @Start = GETDATE()
SELECT @NumElements = COUNT(*) FROM fn_split_big(@Array,',')
PRINT 'UnConventional: ' + CAST(DATEDIFF(second,@Start,GETDATE()) AS varchar)
SET @Start = GETDATE()
SELECT @NumElements = COUNT(*) FROM fn_split_cte(@Array,',')
PRINT 'Recursive CTE: ' + CAST(DATEDIFF(second,@Start,GETDATE()) AS varchar)

